/*
   arguments.c
   
   THIS SOFTWARE IS IN PUBLIC DOMAIN
     There is no limitation on any kind of usage of this code. If you
     modify this code please remove this header and rename the file.
     
     You use it because it is yours.
     
     Originally written and maintained by Matlib (matlib@laser.t28.net),
     any comments or suggestions are welcome.
     
   Change log:
        1  2007-03-05  Matlib
        2  2009-02-19  Matlib (minor corrections)
*/



#include "arguments.h"


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>





#ifndef __USE_GNU
  char *program_invocation_short_name = NULL;
#endif

void arguments_error_freeform (char *format, ...)
{
  va_list ap;
#ifdef __USE_GNU
  extern char *program_invocation_short_name;
#endif
  
  va_start (ap, format);
  vfprintf (stderr, format, ap);
  va_end (ap);
  fprintf (stderr, "\nUse %s --help to see the list of available options.\n",
      program_invocation_short_name);
  exit (EXIT_FAILURE);
}

void arguments_error (struct Argument *arg, char *string)
{
  fprintf (stderr, "Option ");
  if (((*arg).lng != NULL) && ((*arg).shrt) != 0)
    fprintf (stderr, "--%s (-%c) ", (*arg).lng, (*arg).shrt);
  else if ((*arg).lng != NULL)
    fprintf (stderr, "--%s ", (*arg).lng);
  else
    fprintf (stderr, "-%c ", (*arg).shrt);
  arguments_error_freeform ("%s", string);
}







ARGUMENT_CALLBACK (argument_bool)
{
  (*arg).value ++;
}

ARGUMENT_CALLBACK (argument_string)
{
  (*arg).value = parameter;
}

ARGUMENT_CALLBACK (argument_int)
{
  char *tail;
  (*arg).value = (void *) (strtol (parameter, &tail, 0));
  if (*tail != 0)
    arguments_error (arg, "expects integer parameter.");
}





void arguments_version (char *program_name, char *program_version)
{
  if (program_name != NULL)
  {
    printf (program_name);
    if (program_version != NULL)
      printf (" version %s", program_version);
    putchar ('\n');
  }
}

void arguments_help_print_arg (struct Argument *arg, int *f)
{
  if (arg -> shrt != 0)
  {
    printf ("  -%c", (*arg).shrt);
    if (arg -> shrt == 'h')
      *f |= 1; /* flags = 1 (-h) | 2 (--help) | 4 (--version) */
  }
  else
    printf ("    ");
  if (arg -> lng != NULL)
  {
    printf ("  --%s", (*arg).lng);
    if (strcmp (arg -> lng, "help") == 0)
      *f |= 2; /* flags again, see note above */
    else if (strcmp (arg -> lng, "version") == 0)
      *f |= 4;
  }
  if ((arg -> flags & ARGUMENT_F_PAR) != 0)
    printf (" <par>");
  putchar ('\n');
}

void arguments_help (char **argv, struct Argument *arg, char *program_name, 
   char *program_version, char *usage, char *description)
{
  int i, f, nl;
  arguments_version (program_name, program_version);
  
  f = 0;
  printf ("Usage: %s [OPTION]... ", argv [0]);
  if (usage != NULL)
    printf (usage);
  putchar ('\n');
  if (description != NULL)
    printf (description);
  putchar ('\n');
  nl = 0;
  if (arg != NULL)
    for (i = 0; arg [i].cb != NULL; i ++)
      if ((arg [i].flags & ARGUMENT_F_OBL) != 0)
      {
        if (nl == 0)
        {
          putchar ('\n');
          nl ++;
        }
        arguments_help_print_arg (&(arg [i]), &f);
      }
  printf ("\nOptional arguments:\n\n");
  if (arg != NULL)
    for (i = 0; arg [i].cb != NULL; i ++)
      if ((arg [i].flags & ARGUMENT_F_OBL) == 0)
        arguments_help_print_arg (&(arg [i]), &f);

  /* print --help and --version if not redefined */
  if (f != 7)
  {
    putchar ('\n');
    printf ((f & 1) == 0 ? "  -h" : "    ");
    if ((f & 2) == 0)
      printf ("  --help");
    putchar ('\n');
    if ((f & 4) == 0)
      printf ("      --version\n");
  }

  putchar ('\n');   
  exit (EXIT_SUCCESS);
}





void arguments_mark_option (char **argv, struct Argument *arg)
{
  if (((*arg).flags & (ARGUMENT_F_OCC + ARGUMENT_F_ONCE)) 
     == ARGUMENT_F_OCC + ARGUMENT_F_ONCE)
    arguments_error (arg, "specified more than once.");
  (*arg).flags |= ARGUMENT_F_OCC;
}

int arguments_parse (char *program_name, char *program_version, 
   char *usage, char *short_description, int argc, char **argv, 
   struct Argument *arg)
{
  int a, i, pos;
  
  #ifndef __USE_GNU
    if (program_invocation_short_name == NULL)
      program_invocation_short_name = argv [0];
  #endif
  for (a = 1; (a < argc) && (strlen (argv [a]) > 1) && 
     (argv [a][0] == '-'); a ++)
  {
    if (argv [a][1] == '-')     /* --long options */
    {
      if (argv [a][2] == 0) /* stop processing options at -- */
      {
        a ++;
        break;
      }
      for (pos = 0; (argv [a][pos] != 0) && (argv [a][pos] != '='); pos ++);
      i = -1;
      do
      {
        i ++;
        if ((arg == NULL) || (arg [i].cb == NULL))
        {
          if (strcmp (argv [a], "--version") == 0)
          {
            arguments_version (program_name, program_version);
            exit (EXIT_SUCCESS);
          }
          if (strcmp (argv [a], "--help") == 0)
            arguments_help (argv, arg, program_name, program_version, usage, 
               short_description);
          arguments_error_freeform ("Unknown option: %s.", argv [a]);
        }
      }
      while ((arg [i].lng == NULL) 
         || (strncmp (arg [i].lng, argv [a] + 2, pos - 2  ) != 0));
      arguments_mark_option (argv, &(arg [i]));
      if (argv [a][pos] == '=')
      {
        pos ++;
        if ((arg [i].flags & ARGUMENT_F_PAR) == 0)
          arguments_error (&(arg [i]), "does not accept any parameters.");
      }
      if ((arg [i].flags & ARGUMENT_F_PAR) == 0)
        arg [i].cb (argc, argv, &(arg [i]), NULL);
    }
    else                        /* -s hort options */
    {
      for (pos = 1; argv [a][pos] != 0; pos ++)
      {
        i = -1;
        do
        {
          i ++;
          if ((arg == NULL) || (arg [i].cb == NULL))
          {
            if (argv [a][pos] == 'h')
              arguments_help (argv, arg, program_name, program_version, usage, 
                 short_description);
            else
              arguments_error_freeform ("Unknown option: -%c.", argv [a][pos]);
          }
        }
        while (arg [i].shrt != argv [a][pos]);
        arguments_mark_option (argv, &(arg [i]));
        if ((arg [i].flags & ARGUMENT_F_PAR) == 0)
          arg [i].cb (argc, argv, &(arg [i]), NULL);
        else
        {
          pos ++;
          break;
        }
      }
    }
    
    if ((arg [i].flags & ARGUMENT_F_PAR) != 0)
    {
      if (argv [a][pos] == 0)
      {
        a ++;
        pos = 0;
        if (a >= argc)
          arguments_error (&(arg [i]), "requires parameter.");
      }
      arg [i].cb (argc, argv, &(arg [i]), argv [a] + pos);
    }
  }
  
  
  /* Checking which obligatory arguments were not found */
  if (arg != NULL)
    for (i = 0; arg [i].cb != NULL; i ++)
      if (((arg [i].flags & ARGUMENT_F_OBL) != 0) 
         && ((arg [i].flags & ARGUMENT_F_OCC) == 0))
        arguments_error (&(arg [i]), "must be specified.");
  
  return a;
}
