/*
   arguments.h
   
   THIS SOFTWARE IS IN PUBLIC DOMAIN
     There is no limitation on any kind of usage of this code. If you
     modify this code please remove this header and rename the file.
     
     You use it because it is yours.
     
     Originally written and maintained by Matlib (matlib@laser.t28.net),
     any comments or suggestions are welcome.
     
   Change log:
        1  2007-03-05  Matlib

*/



#ifndef ARGUMENTS_H

  #define ARGUMENTS_H YES

  
  
  /* predefinitions */
  struct Argument;



  /* 
     Callback procedure type definition
     Use ARGUMENT_CALLBACK to define a callback to remain compatible
     with possible future versions of this library.
     
     Arguments are:
       * int argc -- same as void main (int argc, char **argv).
       * char **argv -- same as void main (int argc, char **argv).
       * struct Argument *arg -- pointer to the argument structure that 
           matched the command line option; it is not the whole array as 
           passed to arguments_parse!
       * char *parameter -- option's parameter if flag ARGUMENT_F_PAR was 
           set. NULL otherwise.
  */
  typedef void Argument_Callback (int, char **, struct Argument *, char *);
  #define ARGUMENT_CALLBACK(_NAME_) \
    void _NAME_ (int argc, char **argv, struct Argument *arg, char *parameter)



  /* struct Argument
       * char shrt - short option (one character), 0 if none.
       * char *lng - long option (string), NULL if none (at least one of
           the two must be present!).
       * Argument_Callback *cb - the procedure that is called every
           time the option is found.
       * int flags - flags, see section Flags.
       * void *value - place where argument callback may store the result. */
  struct Argument 
  {
    char shrt;
    char *lng;
    Argument_Callback *cb;
    int flags;
    void *value;
  };
  
  /* Flags (struct Argument.flags): */
  #define ARGUMENT_F_OCC    0x01    /* argument found in command line */
  #define ARGUMENT_F_PAR    0x02    /* argument requires parameter */
  #define ARGUMENT_F_ONCE   0x04    /* argument may occur only once */
  #define ARGUMENT_F_OBL    0x08    /* argument must occur at least once */
  
  

  
  

  /* Standard callbacks */
  
  /* argument_bool increases struct Argument.value by 1. Without any flags
     it is verbosity-like option (use -vvvv to increase verbosity). With
     ARGUMENT_F_ONCE it is boolean type option (-x to turn on) */
  ARGUMENT_CALLBACK (argument_bool);
  
  /* argument_string points struct Argument.value at argument's parameter.
     Should be used in with flags ARGUMENT_F_PAR | ARGUMENT_F_ONCE to
     make any sense */
  ARGUMENT_CALLBACK (argument_string);
  
  /* argument_int parses the integer in parameter and stores the result in
     struct Argument.value */
  ARGUMENT_CALLBACK (argument_int);

  
  
  
  /* arguments_parse - main procedure; parses the command line options and
     executes corresponding callbacks. 
     
     Parameters are:
       * program_name - your program's name shown in --version.
       * program_version - version, shown in --version.
       * usage - the usage string shows in --help in the form:
         Usage: _program_ [OPTIONS]... _usage_
       * short_description - shown in --help below usage.
       * int argc - as in int main (int argc, char **argv)
       * char **argv - as in int main (int argc, char **argv)
       * struct Argument *arg - the array of available options */
  int arguments_parse (char *program_name, char *program_version, 
     char *usage, char *short_description, int argc, char **argv, 
     struct Argument *arg);
     
     
     
  /* arguments_error_freeform - basic template to print errors,
        exits with rc == EXIT_FAILURE
        
     Parameters are:
       * char *format - format/template, same as printf,
  */
  void arguments_error_freeform (char *format, ...);
  
  
  
  /* arguments_error - report an erroneous usage of command line arguments.
  
     Parameters are:
       * struct Argument *arg - pointer to the argument that caused the
         error, not the whole array as in arguments_parse!
       * char *string - informational string appended to error message in
         the form:
         Argument _arg_ _string_
         Use _program_ --help to see...  */
  void arguments_error (struct Argument *arg, char *string);


#endif
